/*************************************************************
 *
 *  Copyright (c) 2017 The MathJax Consortium
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * @fileoverview  A visitor that produces a serilaied MathML string
 *                (replacement for toMathML() output from v2)
 *
 * @author dpvc@mathjax.org (Davide Cervone)
 */

import {MmlVisitor} from './MmlVisitor.js';
import {MmlFactory} from './MmlFactory.js';
import {MmlNode, TextNode, XMLNode} from './MmlNode.js';

/*****************************************************************/
/*
 *  Implements the SerializedMmlVisitor (subclass of MmlVisitor)
 */

export class SerializedMmlVisitor extends MmlVisitor {

    /*
     * Convert the tree rooted at a particular node into a serialized MathML string
     *
     * @param {MmlNode} node  The node to use as the root of the tree to traverse
     * @return {string}  The MathML string representing the internal tree
     */
    public visitTree(node: MmlNode) {
        return this.visitNode(node, '');
    }

    /*
     * @param {TextNode} node  The text node to visit
     * @param {string} space  The amount of indenting for this node
     * @return {string}  The text of the node
     */
    public visitTextNode(node: TextNode, space: string) {
        return node.getText();
    }

    /*
     * @param {XMLNode} node  The XML node to visit
     * @param {string} space  The amount of indenting for this node
     * @return {string}  The serialziation of the XML node (not implemented yet).
     */
    public visitXMLNode(node: XMLNode, space: string) {
        return '[XML Node not implemented]';
    }

    /*
     * Vist an inferred mrow, but don't add the inferred row itself (since
     * it is supposed to be inferred).
     *
     * @param {MmlNode} node  The inferred mrow to visit
     * @param {string} space  The amount of indenting for this node
     * @return {string}  The serialzied contents of the mrow, properly indented
     */
    public visitInferredMrowNode(node: MmlNode, space: string) {
        let mml = [];
        for (const child of node.childNodes) {
            mml.push(this.visitNode(child, space));
        }
        return mml.join('\n');
    }

    /*
     * The generic visiting function:
     *   Make the string versino of the open tag, properly indented, with it attributes
     *   Increate the indentation level
     *   Add the childnodes
     *   Add the end tag with proper spacing (empty tags have the close tag following directly)
     *
     * @param {MmlNode} node  The node to visit
     * @param {Element} parent  The DOM parent to which this node should be added
     */
    public visitDefault(node: MmlNode, space: string) {
        let kind = node.kind;
        let [nl, endspace] = (node.isToken || node.childNodes.length === 0 ? ['', ''] : ['\n', space]);
        let mml = space + '<' + kind + this.getAttributes(node) + '>' + nl;
        space += '  ';
        for (const child of node.childNodes) {
            mml += this.visitNode(child, space) + nl;
        }
        mml += endspace + '</' + kind + '>';
        return mml;
    }

    /*
     * @param {MmlNode} node  The node whose attributes are to be produced
     * @return {string}  The attribute list as a string
     */
    protected getAttributes(node: MmlNode) {
        let ATTR = '';
        let attributes = node.attributes.getAllAttributes();
        for (const name of Object.keys(attributes)) {
            ATTR += ' ' + name + '="' + this.quoteAttribute(attributes[name].toString()) + '"';
        }
        return ATTR;
    }

    /*
     *  Convert HTML special characters to entities (&amp;, &lt;, &gt;, &quot;)
     *  Convert multi-character Unicode characters to entities
     *  Convert non-ASCII characters to entities.
     *
     * @param {string} value  The string to be made HTML escaped
     * @return {string}  The string with escaping performed
     */
    protected quoteAttribute(value: string) {
        return value
            .replace(/&/g, '&amp;')
            .replace(/</g, '&lt;').replace(/>/g, '&gt;')
            .replace(/\"/g, '&quot;')
            .replace(/([\uD800-\uDBFF].)/g, (m, c) => {
                return '&#x' + ((c.charCodeAt(0) - 0xD800) * 0x400 +
                                (c.charCodeAt(1) - 0xDC00) + 0x10000).toString(16).toUpperCase() + ';';
            })
            .replace(/([\u0080-\uD7FF\uE000-\uFFFF])/g, (m, c) => {
                return '&#x' + c.charCodeAt(0).toString(16).toUpperCase() + ';';
            });
    }

}
