"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var MmlNode_js_1 = require("../../core/MmlTree/MmlNode.js");
var TreeHelper_js_1 = require("./TreeHelper.js");
var OldParser_js_1 = require("./OldParser.js");
var ParserUtil;
(function (ParserUtil) {
    var emPerInch = 7.2;
    var pxPerInch = 72;
    var UNIT_CASES = {
        'em': function (m) { return m; },
        'ex': function (m) { return m * .43; },
        'pt': function (m) { return m / 10; },
        'pc': function (m) { return m * 1.2; },
        'px': function (m) { return m * emPerInch / pxPerInch; },
        'in': function (m) { return m * emPerInch; },
        'cm': function (m) { return m * emPerInch / 2.54; },
        'mm': function (m) { return m * emPerInch / 25.4; },
        'mu': function (m) { return m / 18; },
    };
    function matchDimen(dim) {
        return dim.match(/^(-?(?:\.\d+|\d+(?:\.\d*)?))(px|pt|em|ex|mu|pc|in|mm|cm)$/);
    }
    ParserUtil.matchDimen = matchDimen;
    ;
    function dimen2em(dim) {
        var match = matchDimen(dim);
        var m = parseFloat(match[1] || '1'), unit = match[2];
        var func = UNIT_CASES[unit];
        return func ? func(m) : 0;
    }
    ParserUtil.dimen2em = dimen2em;
    ;
    function Em(m) {
        if (Math.abs(m) < .0006) {
            return '0em';
        }
        return m.toFixed(3).replace(/\.?0+$/, '') + 'em';
    }
    ParserUtil.Em = Em;
    ;
    function fenced(open, mml, close) {
        TreeHelper_js_1.TreeHelper.printMethod('fenced');
        var mrow = TreeHelper_js_1.TreeHelper.createNode('mrow', [], { open: open, close: close, texClass: MmlNode_js_1.TEXCLASS.INNER });
        var openNode = TreeHelper_js_1.TreeHelper.createText(open);
        var mo = TreeHelper_js_1.TreeHelper.createNode('mo', [], { fence: true, stretchy: true, symmetric: true, texClass: MmlNode_js_1.TEXCLASS.OPEN }, openNode);
        TreeHelper_js_1.TreeHelper.appendChildren(mrow, [mo]);
        if (TreeHelper_js_1.TreeHelper.isType(mml, 'mrow') && TreeHelper_js_1.TreeHelper.isInferred(mml)) {
            TreeHelper_js_1.TreeHelper.appendChildren(mrow, TreeHelper_js_1.TreeHelper.getChildren(mml));
        }
        else {
            TreeHelper_js_1.TreeHelper.appendChildren(mrow, [mml]);
        }
        var closeNode = TreeHelper_js_1.TreeHelper.createText(close);
        mo = TreeHelper_js_1.TreeHelper.createNode('mo', [], { fence: true, stretchy: true, symmetric: true, texClass: MmlNode_js_1.TEXCLASS.CLOSE }, closeNode);
        TreeHelper_js_1.TreeHelper.appendChildren(mrow, [mo]);
        return mrow;
    }
    ParserUtil.fenced = fenced;
    ;
    function fixedFence(open, mml, close) {
        TreeHelper_js_1.TreeHelper.printMethod('fixedFence');
        var mrow = TreeHelper_js_1.TreeHelper.createNode('mrow', [], { open: open, close: close, texClass: MmlNode_js_1.TEXCLASS.ORD });
        if (open) {
            TreeHelper_js_1.TreeHelper.appendChildren(mrow, [mathPalette(open, 'l')]);
        }
        if (TreeHelper_js_1.TreeHelper.isType(mml, 'mrow')) {
            TreeHelper_js_1.TreeHelper.appendChildren(mrow, TreeHelper_js_1.TreeHelper.getChildren(mml));
        }
        else {
            TreeHelper_js_1.TreeHelper.appendChildren(mrow, [mml]);
        }
        if (close) {
            TreeHelper_js_1.TreeHelper.appendChildren(mrow, [mathPalette(close, 'r')]);
        }
        return mrow;
    }
    ParserUtil.fixedFence = fixedFence;
    ;
    function mathPalette(fence, side) {
        TreeHelper_js_1.TreeHelper.printMethod('mathPalette');
        if (fence === '{' || fence === '}') {
            fence = '\\' + fence;
        }
        var D = '{\\bigg' + side + ' ' + fence + '}';
        var T = '{\\big' + side + ' ' + fence + '}';
        return new OldParser_js_1.OldParser('\\mathchoice' + D + T + T + T, {}, {}).mml();
    }
    ParserUtil.mathPalette = mathPalette;
    ;
    function combineRelations(mml) {
        TreeHelper_js_1.TreeHelper.printMethod('combineRelations: ');
        var m1, m2;
        var children = TreeHelper_js_1.TreeHelper.getChildren(mml);
        for (var i = 0, m = children.length; i < m; i++) {
            if (children[i]) {
                if (TreeHelper_js_1.TreeHelper.isType(mml, 'mrow')) {
                    while (i + 1 < m && (m1 = children[i]) && (m2 = children[i + 1]) &&
                        TreeHelper_js_1.TreeHelper.isType(m1, 'mo') && TreeHelper_js_1.TreeHelper.isType(m2, 'mo') &&
                        TreeHelper_js_1.TreeHelper.getTexClass(m1) === MmlNode_js_1.TEXCLASS.REL &&
                        TreeHelper_js_1.TreeHelper.getTexClass(m2) === MmlNode_js_1.TEXCLASS.REL) {
                        if (TreeHelper_js_1.TreeHelper.getProperty(m1, 'variantForm') === TreeHelper_js_1.TreeHelper.getProperty(m2, 'variantForm') &&
                            TreeHelper_js_1.TreeHelper.getAttribute(m1, 'mathvariant') === TreeHelper_js_1.TreeHelper.getAttribute(m2, 'mathvariant')) {
                            TreeHelper_js_1.TreeHelper.appendChildren(m1, TreeHelper_js_1.TreeHelper.getChildren(m2));
                            children.splice(i + 1, 1);
                            m--;
                        }
                        else {
                            TreeHelper_js_1.TreeHelper.untested('Combine Relations Case 2');
                            TreeHelper_js_1.TreeHelper.setAttribute(m1, 'rspace', '0pt');
                            TreeHelper_js_1.TreeHelper.setAttribute(m2, 'lspace', '0pt');
                            i++;
                        }
                    }
                }
                if (!children[i].isToken) {
                    combineRelations(children[i]);
                }
            }
        }
    }
    ParserUtil.combineRelations = combineRelations;
    ;
    function fixInitialMO(data) {
        TreeHelper_js_1.TreeHelper.printMethod('AMS-fixInitialMO');
        for (var i = 0, m = data.length; i < m; i++) {
            var child = data[i];
            if (child && (!TreeHelper_js_1.TreeHelper.isType(child, 'mspace') &&
                (!TreeHelper_js_1.TreeHelper.isType(child, 'TeXAtom') ||
                    (TreeHelper_js_1.TreeHelper.getChildren(child)[0] &&
                        TreeHelper_js_1.TreeHelper.getChildren(TreeHelper_js_1.TreeHelper.getChildren(child)[0]).length)))) {
                if (TreeHelper_js_1.TreeHelper.isEmbellished(child)) {
                    var mi = TreeHelper_js_1.TreeHelper.createNode('mi', [], {});
                    data.unshift(mi);
                }
                break;
            }
        }
    }
    ParserUtil.fixInitialMO = fixInitialMO;
    ;
})(ParserUtil = exports.ParserUtil || (exports.ParserUtil = {}));
//# sourceMappingURL=ParserUtil.js.map